# flood-fill-filter - edge bunches detection tool.
# Copyright (C) 2020  Georgy Ustinov  <georgy.ustinov.hello@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


class Xyz:
    def __init__(self, yXYZ, xXYZ, zXYZ):
        assert yXYZ.shape[0] == xXYZ.shape[0]
        assert yXYZ.shape[0] == zXYZ.shape[0]
        assert yXYZ.shape[1] == xXYZ.shape[1]
        assert yXYZ.shape[1] == zXYZ.shape[1]
        self.yXYZ = yXYZ
        self.xXYZ = xXYZ
        self.zXYZ = zXYZ
        self.h = yXYZ.shape[0]
        self.w = yXYZ.shape[1]

    def expand(self, margin):
        """Expand with inverted lightness border."""
        invert = lambda array: -array + 1.0
        first_row = lambda array: array[:1]
        last_row = lambda array: array[-1:]

        first_column = lambda array: array[:, :1]
        last_column = lambda array: array[:, -1:]

        y = np.concatenate(
            (
                np.repeat(invert(first_row(self.yXYZ)), margin, axis=0),
                self.yXYZ,
                np.repeat(invert(last_row(self.yXYZ)), margin, axis=0)
            ))

        x = np.concatenate(
            (
                np.repeat(first_row(self.xXYZ), margin, axis=0),
                self.xXYZ,
                np.repeat(last_row(self.xXYZ), margin, axis=0)
            ))

        z = np.concatenate(
            (
                np.repeat(first_row(self.zXYZ), margin, axis=0),
                self.zXYZ,
                np.repeat(last_row(self.zXYZ), margin, axis=0)
            ))

        y = np.concatenate(
            (
                np.repeat(invert(first_column(y)), margin, axis=1),
                y,
                np.repeat(invert(last_column(y)), margin, axis=1)
            ),
            axis=1
        )

        x = np.concatenate(
            (
                np.repeat(first_column(x), margin, axis=1),
                x,
                np.repeat(last_column(x), margin, axis=1)
            ),
            axis=1
        )

        z = np.concatenate(
            (
                np.repeat(first_column(z), margin, axis=1),
                z,
                np.repeat(last_column(z), margin, axis=1)
            ),
            axis=1
        )

        return Xyz(y, x, z)

    def shape(self):
        return self.yXYZ.shape

    def left(self, n=1):
        if 0 == n:
            return self
        elif n < 0:
            return self.right(-n)
        return Xyz(self.yXYZ[:, :-n], self.xXYZ[:, :-n], self.zXYZ[:, :-n])

    def right(self, n=1):
        if 0 == n:
            return self
        elif n < 0:
            return self.left(-n)
        return Xyz(self.yXYZ[:, n:], self.xXYZ[:, n:], self.zXYZ[:, n:])

    def top(self, n=1):
        if 0 == n:
            return self
        elif n < 0:
            return self.bottom(-n)
        return Xyz(self.yXYZ[:-n, :], self.xXYZ[:-n, :], self.zXYZ[:-n, :])

    def bottom(self, n=1):
        if 0 == n:
            return self
        elif n < 0:
            return self.top(-n)
        return Xyz(self.yXYZ[n:, :], self.xXYZ[n:, :], self.zXYZ[n:, :])

    def eq(self, other, y_threshold):
        chroma_threshold_factor = 4

        y_equal = np.abs(other.yXYZ - self.yXYZ) < y_threshold

        xd = np.abs(other.xXYZ - self.xXYZ)
        zd = np.abs(other.zXYZ - self.zXYZ)

        tf = y_threshold * chroma_threshold_factor

        return np.logical_and(
            y_equal,
            np.logical_or(
                np.logical_or(self.yXYZ <= (25 / 255), other.yXYZ <= (25 / 255)),
                np.logical_or(
                    np.logical_and(
                        np.logical_and(
                            np.logical_or(self.yXYZ > 0.5, other.yXYZ > 0.5),
                            xd < tf * 2
                        ),
                        zd < tf * 2
                    ),
                    np.logical_and(
                        xd < tf,
                        zd < tf
                    )
                )
            )
        )



# flood-fill-filter - edge bunches detection tool.
# Copyright (C) 2020  Georgy Ustinov  <georgy.ustinov.hello@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

__all__ = (
    'to_linear',
    'from_linear'
)


def srgb_to_linear(cb):
    c = cb / 255.0
    a = 0.055
    array_if = c / 12.92
    array_else = np.power((c + a) / (1 + a), 2.4)
    return np.where(c <= 0.04045, array_if, array_else)


def float_to_byte(a):
    return np.fmax(np.fmin(np.rint(255.0 * a), 255.0), 0.0)


def linear_to_srgb_gamma_correction(lin):
    a = 0.055
    array_if = lin * 12.92
    array_else = np.power(lin, 1.0 / 2.4) * (1 + a) - a
    return np.where(lin <= 0.0031308, array_if, array_else)


def linear_to_srgb(lin):
    return float_to_byte(linear_to_srgb_gamma_correction(lin))


def to_linear(r, g, b, a):
    result = np.zeros((r.shape[0], r.shape[1], 4), dtype=np.single)
    result[:, :, 0] = srgb_to_linear(r)
    result[:, :, 1] = srgb_to_linear(g)
    result[:, :, 2] = srgb_to_linear(b)
    result[:, :, 3] = a / 255.0
    return result


def from_linear(linear_rgba):
    lr = linear_rgba[:, :, 0]
    lg = linear_rgba[:, :, 1]
    lb = linear_rgba[:, :, 2]
    la = linear_rgba[:, :, 3]
    result = np.zeros((lr.shape[0], lr.shape[1], 4), dtype=np.single)
    result[:, :, 0] = linear_to_srgb(lr)
    result[:, :, 1] = linear_to_srgb(lg)
    result[:, :, 2] = linear_to_srgb(lb)
    result[:, :, 3] = float_to_byte(la)
    return result



# flood-fill-filter - edge bunches detection tool.
# Copyright (C) 2020  Georgy Ustinov  <georgy.ustinov.hello@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np


def equality_matrices(original_image, kernel_margin, y_threshold, offsets, pack=True):

    assert np.min(original_image.yXYZ) >= 0.0
    assert np.min(original_image.xXYZ) >= 0.0
    assert np.min(original_image.zXYZ) >= 0.0

    assert np.max(original_image.yXYZ) <= 1.0
    assert np.max(original_image.xXYZ) <= 1.0
    assert np.max(original_image.zXYZ) <= 1.089

    image = original_image.expand(kernel_margin)

    result = np.zeros(
        (
            original_image.h,
            original_image.w,
            len(offsets)
        ),
        dtype=np.bool
    )

    for offset_index in range(len(offsets)):
        y_offset, x_offset = offsets[offset_index]

        if (0 == y_offset) and (0 == x_offset):
            continue

        data_to_compare = image

        if 0 == y_offset:
            data_to_compare = data_to_compare.top(kernel_margin).bottom(kernel_margin)
        elif y_offset < 0:
            data_to_compare = data_to_compare \
                .bottom(kernel_margin + abs(y_offset)) \
                .top(kernel_margin - abs(y_offset))
        else:
            data_to_compare = data_to_compare \
                .top(kernel_margin + y_offset) \
                .bottom(kernel_margin - y_offset)

        if 0 == x_offset:
            data_to_compare = data_to_compare.left(kernel_margin).right(kernel_margin)
        elif x_offset < 0:
            data_to_compare = data_to_compare \
                .right(kernel_margin + abs(x_offset)) \
                .left(kernel_margin - abs(x_offset))
        else:
            data_to_compare = data_to_compare \
                .left(kernel_margin + x_offset) \
                .right(kernel_margin - x_offset)

        result[:, :, offset_index] = \
            data_to_compare.eq(original_image, y_threshold)

    if pack:
        return np.packbits(result, axis=2)
    else:
        return result



# flood-fill-filter - edge bunches detection tool.
# Copyright (C) 2020  Georgy Ustinov  <georgy.ustinov.hello@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import unittest

import os

from flood_fill_filter.flood import read_linear, AdjacentMatrixHolder
import flood_fill_filter.private.xyz_loader as xyz_loader

from flood_fill_filter.private.calculations import *

class TestCalculations(unittest.TestCase):
    def test_equality_matrices(self):
        directory = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir, 'samples3')
        input_image = 'xm20.bmp'

        linear_rgba = read_linear(os.path.join(directory, input_image))
        xyz = xyz_loader.from_rgba(linear_rgba)

        kernel_margin = 4
        adjacency_martix_holder = AdjacentMatrixHolder(kernel_margin)

        equality_masks = equality_matrices(xyz, kernel_margin, 0.08, adjacency_martix_holder.offsets, pack=False)

        kernel_diameter = kernel_margin + 1 + kernel_margin

        assert equality_masks.shape[0] == 136
        assert equality_masks.shape[1] == 195
        assert equality_masks.shape[2] == kernel_diameter * kernel_diameter

        csv_offsets = np.loadtxt(os.path.join(directory, 'xm20_eq_offsets.csv'), delimiter=',')

        expected = np.loadtxt(os.path.join(directory, 'xm20_eq.csv.gz'), delimiter=',')
        expected = expected.reshape((136, 195, kernel_diameter * kernel_diameter))

        for csv_offset_index in range(csv_offsets.shape[0]):
            csv_offset = tuple(csv_offsets[csv_offset_index])
            holder_offset_index = adjacency_martix_holder.offsets.index(csv_offset)
            assert np.allclose(equality_masks[:, :, holder_offset_index], expected[:, :, csv_offset_index])



# flood-fill-filter - edge bunches detection tool.
# Copyright (C) 2020  Georgy Ustinov  <georgy.ustinov.hello@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from flood_fill_filter.private.xyz import Xyz
from flood_fill_filter.linear_srgb_conv import linear_to_srgb_gamma_correction


def white_lowering_function(gamma_corrected_luma):
    """
    White is not very important for this filter.
    """
    t = 0.7
    if gamma_corrected_luma > t:
        return gamma_corrected_luma - gamma_corrected_luma * 0.25 * ((gamma_corrected_luma - t) / (1 - t))
    else:
        return gamma_corrected_luma


v_white_lowering_function = \
    np.vectorize(white_lowering_function)


def from_rgba(linear_rgba):
    r = linear_rgba[:, :, 0]
    g = linear_rgba[:, :, 1]
    b = linear_rgba[:, :, 2]

    y = r * 0.2126 + g * 0.7152 + b * 0.0722
    x = r * 0.4124 + g * 0.3576 + b * 0.1805
    z = r * 0.0193 + g * 0.1192 + b * 0.9505

    y_gamma = linear_to_srgb_gamma_correction(y)
    y_gamma_low_white = v_white_lowering_function(y_gamma)

    return Xyz(y_gamma_low_white, x, z)



# flood-fill-filter - edge bunches detection tool.
# Copyright (C) 2020  Georgy Ustinov  <georgy.ustinov.hello@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import os

import math
from multiprocessing import Pool

import numpy as np
from PIL import Image

import flood_fill_filter.private.calculations as calculations
import flood_fill_filter.private.xyz_loader as xyz_loader
from flood_fill_filter.linear_srgb_conv import *
from flood_fill_filter.private.xyz import Xyz

import functools

__all__ = (
    'read_linear',
    'filter',
    'to_8_bit',
    'save'
)

builtin_filter = filter


def read_linear(file):
    im = Image.open(file)
    return pil_image_to_linear(im)


def pil_image_to_linear(image):
    rgba = image

    if 'RGBA' != rgba.mode:
        rgba = rgba.convert('RGBA')

    im = np.array(rgba, dtype=np.float32)
    return to_linear(im[:, :, 0], im[:, :, 1], im[:, :, 2], im[:, :, 3])


@functools.lru_cache(maxsize=5000)
def lrtb(coordinate, size, kernel_margin):
    a = max(0, coordinate - kernel_margin)
    b = min(size - 1, coordinate + kernel_margin)
    return (1 + b - a)


def connect_offsets(adjacency_martix, p1, p2):
    adjacency_martix[p1, p2] = True
    adjacency_martix[p2, p1] = True


class AdjacentMatrixHolder:
    def __init__(self, kernel_margin, prototype=None):
        if prototype:
            self.kernel_margin = prototype.kernel_margin
            self.matrix = prototype.matrix.copy()
            self.offsets = prototype.offsets.copy()
        else:
            self.kernel_margin = kernel_margin
            kernel_diameter = kernel_margin + 1 + kernel_margin
            offsets = []

            for yo in range(-kernel_margin, kernel_margin + 1):
                for xo in range(-kernel_margin, kernel_margin + 1):
                    offsets.append((yo, xo))

            assert len(offsets) == (kernel_diameter * kernel_diameter)

            origin = int((kernel_diameter * kernel_diameter) / 2)

            assert offsets[origin] == (0, 0)
            assert offsets[origin - kernel_diameter] == (-1, 0)
            assert offsets[origin + kernel_diameter] == (1, 0)
            assert offsets[origin - 1] == (0, -1)
            assert offsets[origin + 1] == (0, 1)

            adjacency_martix = np.zeros((len(offsets), len(offsets)), dtype=np.bool)

            for yo in range(-kernel_margin, kernel_margin):
                for xo in range(-kernel_margin, kernel_margin):
                    left_top = origin + yo * kernel_diameter + xo
                    right_top = left_top + 1
                    left_bottom = left_top + kernel_diameter
                    right_bottom = left_bottom + 1
                    connect_offsets(adjacency_martix, left_top, right_top)
                    connect_offsets(adjacency_martix, left_top, left_bottom)
                    connect_offsets(adjacency_martix, right_top, right_bottom)
                    connect_offsets(adjacency_martix, left_bottom, right_bottom)

            self.matrix = np.packbits(adjacency_martix, axis=1)
            self.offsets = offsets

        self.offsets_origin = int(len(self.offsets) / 2)

        self.not_offsets_origin_mask = np.ones((len(self.offsets)), dtype=np.bool)
        self.not_offsets_origin_mask[self.offsets_origin] = False
        self.offsets_origin_mask = np.logical_not(self.not_offsets_origin_mask)

        self.not_offsets_origin_mask = int(np.packbits(self.not_offsets_origin_mask).data.hex(), 16)
        self.offsets_origin_mask = int(np.packbits(self.offsets_origin_mask).data.hex(), 16)

    @functools.lru_cache(maxsize=64000)
    def get_or_result(self, this_step_result):
        this_step_result_bytes = this_step_result.to_bytes(math.ceil(len(self.offsets) / 8), byteorder='big')
        this_step_result_array = np.frombuffer(this_step_result_bytes, dtype=np.uint8)
        this_step_result_indices = np.nonzero(np.unpackbits(this_step_result_array))[0].tolist()
        adjacent_offset_rows = self.matrix[this_step_result_indices]
        return int(np.bitwise_or.reduce(adjacent_offset_rows, axis=0).data.hex(), 16)

    def get_offsets_origin_mask(self):
        return self.offsets_origin_mask

    def get_not_offsets_origin_mask(self):
        return self.not_offsets_origin_mask


def filter_chunk(input):
    equality_matrices_rows, h, w, min_y, max_y_exclusive, kernel_margin, \
    adjacency_martix_holder, ratio_threshold = input

    result = np.zeros((max_y_exclusive - min_y, w), dtype=np.bool)

    origin_mask = adjacency_martix_holder.get_offsets_origin_mask()
    not_origin_mask = adjacency_martix_holder.get_not_offsets_origin_mask()

    for y in range(min_y, max_y_exclusive):
        yi = y - min_y

        tb = lrtb(y, h, kernel_margin)

        for x in range(w):
            count_threshold = int((lrtb(x, w, kernel_margin) * tb - 1) * ratio_threshold)

            filled_offsets = 0x0
            filled_offsets_count = 0

            previous_filling_iteration_result = origin_mask

            get_eq_mask = int(equality_matrices_rows[yi, x].data.hex(), 16)

            while (filled_offsets_count <= count_threshold) and (0x0 != previous_filling_iteration_result):
                or_result = adjacency_martix_holder.get_or_result(previous_filling_iteration_result)
                what_to_fill_next = or_result & ~(not_origin_mask & filled_offsets)

                previous_filling_iteration_result = what_to_fill_next & get_eq_mask
                filled_offsets = filled_offsets | previous_filling_iteration_result
                filled_offsets_count = bin(filled_offsets).count("1")

            result[yi, x] = filled_offsets_count > count_threshold

    return {
        'min_y': min_y,
        'max_y_exclusive': max_y_exclusive,
        'array': result
    }


def filter(linear_rgba, y_threshold=0.08, kernel_margin=4, ratio_threshold=0.45, denoise=False, single_thread=False):
    original_image = xyz_loader.from_rgba(linear_rgba)
    first_pass = one_pass(original_image, y_threshold, kernel_margin, ratio_threshold, single_thread).astype(np.float32)

    if denoise:
        first_pass_xyz = Xyz(first_pass, np.zeros_like(first_pass), np.zeros_like(first_pass))
        second_pass = np.logical_not(one_pass(first_pass_xyz, y_threshold=0.08, kernel_margin=4, ratio_threshold=0.05,
                                              single_thread=single_thread))
        return np.logical_or(first_pass, second_pass)
    else:
        return first_pass


def one_pass(original_image, y_threshold, kernel_margin, ratio_threshold, single_thread):
    assert ratio_threshold > 0.0
    assert ratio_threshold < 1.0

    adjacency_martix_holder = AdjacentMatrixHolder(kernel_margin)

    equality_masks = calculations.equality_matrices(original_image, kernel_margin, y_threshold,
                                                    adjacency_martix_holder.offsets)

    flood_fill_result = np.zeros((original_image.h, original_image.w), dtype=np.bool)

    h, w = original_image.h, original_image.w

    worker_count = os.cpu_count()
    if worker_count > 8:
        worker_count = worker_count - 2
    elif worker_count > 4:
        worker_count = worker_count - 1

    if single_thread:
        chunk_size = h
        chunk_count = 1
    else:
        chunk_size = int(h / min(worker_count, h))
        chunk_count = math.ceil(h / chunk_size)

    input_rows = [
        (
            equality_masks[(chunk * chunk_size):(min(h, ((chunk + 1) * chunk_size)))].copy(),
            h, w,
            (chunk * chunk_size),
            (min(h, ((chunk + 1) * chunk_size))),
            kernel_margin,
            AdjacentMatrixHolder(0, adjacency_martix_holder),
            ratio_threshold
        )
        for chunk in range(chunk_count)
    ]

    if single_thread:
        filled_rows = list(map(filter_chunk, input_rows))
    else:
        pool = Pool(worker_count)
        filled_rows = pool.imap_unordered(filter_chunk, input_rows)
        pool.close()

    for filled_row in filled_rows:
        flood_fill_result[filled_row['min_y']:filled_row['max_y_exclusive'], :] = filled_row['array']

    return flood_fill_result


def to_8_bit(a):
    return np.clip(a, 0, 255).astype(np.uint8)


def save(grayscale_numpy_image, filename):
    img = Image.fromarray(grayscale_numpy_image, 'L')
    img.save(filename)



# flood-fill-filter - edge bunches detection tool.
# Copyright (C) 2020  Georgy Ustinov  <georgy.ustinov.hello@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import unittest
import pytest

import numpy as np
import math
from PIL import Image

import flood_fill_filter.flood as flood


@pytest.fixture(scope="class")
def cli_params_class(pytestconfig, request):
    request.cls.single_thread = pytestconfig.getoption("--single-thread")


def load_folder(folder_name, y_threshold, denoise=False, single_thread=False, filename_predicate=lambda s: True):
    directory = os.path.join(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir), folder_name)
    input_list = sorted([f for f in os.listdir(directory) if
                         (f.endswith('_orig.bmp')) and filename_predicate(f)])
    output_list = sorted([f for f in os.listdir(directory) if f.endswith('_fff.png') and filename_predicate(f)])

    diff_list = []

    for i, input_image in enumerate(input_list):
        input = flood.read_linear(os.path.join(directory, input_image))
        output = flood.filter(input, y_threshold=y_threshold, denoise=denoise,
                              single_thread=single_thread)

        expected_output_filename = output_list[i]
        expected_output = np.array(Image.open(os.path.join(directory, expected_output_filename)).convert('L')) > 128
        diff_count = np.sum(np.logical_xor(output, expected_output))

        diff_list.append({
            'file': input_image,
            'output_file': expected_output_filename,
            'shape': input.shape,
            'diff_count': diff_count
        })

    return diff_list


def load_samples3(single_thread):
    directory = os.path.join(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir), 'samples3')
    input_list = ['xm50.bmp', 'xm20.bmp']
    output_list = ['xm50_fff_denoise.png', 'xm20_fff_denoise.png']

    diff_list = []

    for i, input_image in enumerate(input_list):
        input = flood.read_linear(os.path.join(directory, input_image))
        ouput = flood.filter(input, y_threshold=0.08, kernel_margin=4, ratio_threshold=0.45, denoise=True,
                             single_thread=single_thread)

        expected_output_filename = output_list[i]
        expected_output = np.array(Image.open(os.path.join(directory, expected_output_filename)).convert('L')) > 128
        diff_count = np.sum(np.logical_xor(ouput, expected_output))

        diff_list.append({
            'file': input_image,
            'output_file': expected_output_filename,
            'shape': input.shape,
            'diff_count': diff_count
        })

    return diff_list


@pytest.mark.usefixtures("cli_params_class")
class TestSamples(unittest.TestCase):

    @classmethod
    def initialize_samples(cls, o):
        if not hasattr(cls, "samples"):
            assert hasattr(o, "single_thread")
            cls.samples = load_folder('samples', y_threshold=0.092, single_thread=o.single_thread)

    def test_adjacent_matrix_holder_2(self):
        holder = flood.AdjacentMatrixHolder(2)
        assert len(holder.offsets) == 25

        for offset1_index, offset1 in enumerate(holder.offsets):
            request_mask = np.zeros((len(holder.offsets)), dtype=np.bool)
            request_mask[offset1_index] = True
            request_mask = int(np.packbits(request_mask).data.hex(), 16)

            adjacent_offsets = holder.get_or_result(request_mask)
            adjacent_offsets_bytes = adjacent_offsets.to_bytes(math.ceil(len(holder.offsets) / 8), byteorder='big')
            adjacent_offsets_array = np.unpackbits(np.frombuffer(adjacent_offsets_bytes, dtype=np.uint8))

            for offset2_index, offset2 in enumerate(holder.offsets):
                if ((abs(offset1[0] - offset2[0]) == 1) and (offset1[1] == offset2[1])) or \
                        ((abs(offset1[1] - offset2[1]) == 1) and (offset1[0] == offset2[0])):
                    assert adjacent_offsets_array[offset2_index]
                else:
                    assert False == adjacent_offsets_array[offset2_index]

        origin_mask_bytes = holder.get_offsets_origin_mask(). \
            to_bytes(math.ceil(len(holder.offsets) / 8), byteorder='big')
        origin_mask_array = np.unpackbits(np.frombuffer(origin_mask_bytes, dtype=np.uint8))

        not_origin_mask_bytes = holder.get_not_offsets_origin_mask(). \
            to_bytes(math.ceil(len(holder.offsets) / 8), byteorder='big')
        not_origin_mask_array = np.unpackbits(np.frombuffer(not_origin_mask_bytes, dtype=np.uint8))

        assert np.count_nonzero(origin_mask_array) == 1
        assert np.count_nonzero(not_origin_mask_array) == (len(holder.offsets) - 1)

    def test_adjacent_matrix_holder_5(self):
        holder = flood.AdjacentMatrixHolder(5)
        assert len(holder.offsets) == 121

        for offset1_index, offset1 in enumerate(holder.offsets):
            request_mask = np.zeros((len(holder.offsets)), dtype=np.bool)
            request_mask[offset1_index] = True
            request_mask = int(np.packbits(request_mask).data.hex(), 16)

            adjacent_offsets = holder.get_or_result(request_mask)
            adjacent_offsets_bytes = adjacent_offsets.to_bytes(math.ceil(len(holder.offsets) / 8), byteorder='big')
            adjacent_offsets_array = np.unpackbits(np.frombuffer(adjacent_offsets_bytes, dtype=np.uint8))

            for offset2_index, offset2 in enumerate(holder.offsets):
                if ((abs(offset1[0] - offset2[0]) == 1) and (offset1[1] == offset2[1])) or \
                        ((abs(offset1[1] - offset2[1]) == 1) and (offset1[0] == offset2[0])):
                    assert adjacent_offsets_array[offset2_index]
                else:
                    assert False == adjacent_offsets_array[offset2_index]

        origin_mask_bytes = holder.get_offsets_origin_mask(). \
            to_bytes(math.ceil(len(holder.offsets) / 8), byteorder='big')
        origin_mask_array = np.unpackbits(np.frombuffer(origin_mask_bytes, dtype=np.uint8))

        not_origin_mask_bytes = holder.get_not_offsets_origin_mask(). \
            to_bytes(math.ceil(len(holder.offsets) / 8), byteorder='big')
        not_origin_mask_array = np.unpackbits(np.frombuffer(not_origin_mask_bytes, dtype=np.uint8))

        assert np.count_nonzero(origin_mask_array) == 1
        assert np.count_nonzero(not_origin_mask_array) == (len(holder.offsets) - 1)

    def test_90(self):
        self.initialize_samples(self)
        for image in self.samples:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent < (100 - 90), \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_99(self):
        self.initialize_samples(self)
        for image in self.samples:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent < (100 - 99), \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_999(self):
        self.initialize_samples(self)
        for image in self.samples:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent < (100 - 99.9), \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_99_plus_one(self):
        self.initialize_samples(self)
        for image in self.samples:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert 0.0 == diff_per_cent, \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_2_q20(self):
        assert hasattr(self, "single_thread")
        samples2 = load_folder('samples2', y_threshold=0.08, denoise=True,
                               single_thread=self.single_thread,
                               filename_predicate=lambda f: '_q20_' in f)
        for image in samples2:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent == 0, \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_2_q40(self):
        assert hasattr(self, "single_thread")
        samples2 = load_folder('samples2', y_threshold=0.08, denoise=True,
                               single_thread=self.single_thread,
                               filename_predicate=lambda f: '_q40_' in f)
        for image in samples2:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent == 0, \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_2_q70(self):
        assert hasattr(self, "single_thread")
        samples2 = load_folder('samples2', y_threshold=0.08, denoise=True,
                               single_thread=self.single_thread,
                               filename_predicate=lambda f: '_q70_' in f)
        for image in samples2:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent == 0, \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_2_q100(self):
        assert hasattr(self, "single_thread")
        samples2 = load_folder('samples2', y_threshold=0.08, denoise=True,
                               single_thread=self.single_thread,
                               filename_predicate=lambda f: '_q100_' in f)
        for image in samples2:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent == 0, \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )

    def test_denoise(self):
        assert hasattr(self, "single_thread")
        denoise_samples = load_samples3(single_thread=self.single_thread)
        for image in denoise_samples:
            diff_per_cent = image['diff_count'] / (image['shape'][0] * image['shape'][1]) * 100
            assert diff_per_cent == 0, \
                '{} {} has {} different pixels ({}%)'.format(
                    image['file'],
                    image['output_file'],
                    image['diff_count'],
                    diff_per_cent
                )



import setuptools

setuptools.setup(
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    python_requires=">=3.3"
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import sys

import flood_fill_filter.flood as flood

USAGE_EPILOG = '''Example:

    flood_fill_filter input.jpg output.png
    flood_fill_filter -r 5 input.jpg output.png
    flood_fill_filter -d 0.05 input.jpg output.png
    flood_fill_filter -r 13 -a 0.05 input.jpg output.pcx
    flood_fill_filter --denoise input.jpg output.png

It is recommended to leave the optional parameters at their default values.
'''


def float_from_0_to_1_exclusive(arg):
    try:
        f = float(arg)
    except ValueError as e:
        raise argparse.ArgumentTypeError(str(e))

    if f <= 0.0:
        raise argparse.ArgumentTypeError("Must be greater than zero.")
    elif 1.0 <= f:
        raise argparse.ArgumentTypeError("Must be less than one.")

    return f

def radius_type(arg):
    try:
        i = int(arg)
    except ValueError as e:
        raise argparse.ArgumentTypeError(str(e))

    if i < 1:
        raise argparse.ArgumentTypeError("Must be greater than or equal 1.")

    return i

def main():
    parser = argparse.ArgumentParser(
        description='Pixel-accurate JPEG-noise-resistant edge detection algorithm.',
        epilog=USAGE_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument('-d', '--diff', type=float_from_0_to_1_exclusive, default=0.08,
                        help='Y (CIE XYZ) sensitivity. Default: 0.08.')

    parser.add_argument('-a', '--activation-threshold', type=float_from_0_to_1_exclusive, default=0.45,
                        help='The fraction of filled pixels within the fill window needed '
                             'for the white pixel in the output. Default: 0.45.')

    parser.add_argument('-r', '--radius', type=radius_type, default=4,
                        help='The fill window margin. The window width equals 2r+1. Default: 4.')

    parser.add_argument('--denoise', action='store_true',
                        help='Remove free-standing points.')

    parser.add_argument('input')
    parser.add_argument('output')

    namespace = parser.parse_args(sys.argv[1:])

    input = flood.read_linear(namespace.input)
    result = flood.filter(input,
                          denoise=namespace.denoise,
                          y_threshold=namespace.diff,
                          kernel_margin=namespace.radius,
                          ratio_threshold=namespace.activation_threshold)

    flood.save(
        flood.to_8_bit(result * 255),
        namespace.output
    )


if __name__ == "__main__":
    main()
